#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			September 2008
"""
from datetime import datetime
from threading import local
from contextlib import contextmanager
from BTrees.IIBTree import IIBTree
from cocktail.modeling import classgetter
from cocktail.events import event_handler
from cocktail import schema
from cocktail.translations import translations
from cocktail.persistence import PersistentObject, datastore

CHANGES_INDEX_KEY = "woost.models.changesets.ChangeSet.changes-index"

@contextmanager
def changeset_context(author = None):
    """A context manager that eases item versioning. All changes performed on
    CMS items within the call to the manager will be tracked and made part of a
    L{ChangeSet}.

    Calls to this manager can be nested. In that case, only the topmost call
    will actually produce a new changeset, other calls will reuse the root
    changeset.

    @param author: The author of the changeset.
    @type author: L{User<woost.models.User>}

    Example:

        >>> with changeset_context(some_user) as changeset:
        >>>     item1.price = 15.0
        >>>     item1.stock = 3
        >>>     item2.price = 5.75
        >>>     item3 = MyItem()
        >>> len(changeset.changes)
        3
        >>> changeset.changes[item3.id].action
        "create"
        >>> item3.author
        some_user
        >>> changeset.changes[item1.id].changed_members
        set(["price", "stock"])
        >>> changeset.changes[item2.id].changed_members
        set(["price"])
        >>> changeset.changes[item2.id].item_state["price"]
        5.75
    """

    changeset = ChangeSet.current

    # Begin a new changeset
    if changeset is None:
        changeset = ChangeSet()
        changeset.author = author
        changeset.begin()
        changeset.insert()

        try:
            yield changeset
        finally:
            changeset.end()

    # Nested call: don't create a new changeset, just reuse the changeset
    # that is already in place
    else:
        if author and author is not changeset.author:
            raise ValueError("A changeset can only have a single author")

        yield changeset


class ChangeSet(PersistentObject):
    """A persistent record of a set of L{changes<Change>} performed on one or
    more CMS items."""

    members_order = "id", "author", "date", "changes"
    indexed = True
    full_text_indexed = True

    changes = schema.Mapping(
        searchable = False,
        get_item_key = lambda change: change.target and change.target.id
    )

    author = schema.Reference(
        required = True,
        type = "woost.models.User",
        search_control = "cocktail.html.DropdownSelector",
        text_search = True
    )

    date = schema.DateTime(
        required = True,
        default = schema.DynamicDefault(datetime.now)
    )

    _thread_data = local()

    @classgetter
    def current(cls):
        return getattr(cls._thread_data, "current", None)

    @classgetter
    def current_author(cls):
        cs = cls.current
        return cs and cs.author

    def begin(self):

        if self.current:
            raise TypeError("Can't begin a new changeset, another changeset "
                "is already in place")

        self._thread_data.current = self

    def end(self):
        try:
            del self._thread_data.current
        except AttributeError:
            raise TypeError("Can't finalize the current changeset, there's no "
                "changeset in place")

    @classmethod
    def extract_searchable_text(cls, extractor):
        PersistentObject.extract_searchable_text(extractor)
        for change in extractor.current.value.changes.itervalues():
            extractor.extract(change.__class__, change)

    @classgetter
    def changes_index(cls):
        try:
            return datastore.root[CHANGES_INDEX_KEY]
        except KeyError:
            index = IIBTree()
            datastore.root[CHANGES_INDEX_KEY] = index
            return index


class Change(PersistentObject):
    """A persistent record of an action performed on a CMS item."""

    indexed = True
    full_text_indexed = True

    changeset = schema.Reference(
        required = True,
        indexed = True,
        type = "woost.models.ChangeSet"
    )

    action = schema.String(
        required = True,
        indexed = True,
        enumeration = ["create", "modify", "delete"]
    )

    target = schema.Reference(
        required = True,
        type = "woost.models.Item",
        indexed = True,
        bidirectional = True
    )

    changed_members = schema.Collection(
        type = set,
        items = schema.String(),
        indexed = True
    )

    item_state = schema.Mapping(
        required = False
    )

    is_explicit_change = schema.Boolean(
        required = True,
        default = False,
        indexed = True
    )

    def get_previous_change(self):
        target = self.target
        if target is not None:
            changes = target.changes
            index = changes.index(self)
            if index > 0:
                return changes[index - 1]

    def diff(self,
        other_change = None,
        diff_schema = None,
        exclude = None,
        language_subset = None
    ):

        if other_change is None:
            other_change = self.get_previous_change()

        if other_change is None:
            raise ValueError(
                "Can't obtain the diff for %r, it has no parent change"
            )

        if diff_schema is None:
            model = self.target.__class__
            adapter = schema.Adapter()
            adapter.exclude([
                member.name
                for member in model.iter_members()
                if not member.visible or not member.versioned
            ])
            diff_schema = adapter.export_schema(model)

        return schema.diff(
            self.item_state,
            other_change.item_state,
            diff_schema,
            exclude = exclude,
            language_subset = language_subset
        )

    def __translate__(self, language, **kwargs):
        return translations(
            "woost.models.changesets.Change description",
            action = self.action,
            target = self.target
        ) or PersistentObject.__translate__(self, language, **kwargs)

    @event_handler
    def handle_changed(cls, e):
        change = e.source
        member = e.member
        value = e.value

        if member is cls.changeset:
            if e.previous_value is not None and value is not None:
                raise ValueError("Can't move a change between changesets")
            if value is not None and change.is_inserted:
                ChangeSet.changes_index.insert(change.id, value.id)

    @event_handler
    def handle_inserted(cls, e):
        change = e.source
        changeset = change.changeset
        if changeset is not None:
            ChangeSet.changes_index.insert(change.id, changeset.id)

    @event_handler
    def handle_deleting(cls, e):
        try:
            del ChangeSet.changes_index[e.source.id]
        except KeyError:
            pass


class ChangeSetHasChangeExpression(schema.expressions.Expression):

    def __init__(self, target = None, action = None, include_implicit = True):
        self.target = target
        self.action = action
        self.include_implicit = include_implicit

    def op(self, changeset, action):
        for change in changeset.changes.itervalues():
            if (
                # Filter by target
                (
                    self.target is None
                    or (
                        isinstance(change.target, self.target)
                        if isinstance(self.target, type)
                        else change.target is self.target
                    )
                )
                # Filter by action
                and (
                    self.action is None
                    or change.action == self.action
                )
            ):
                return True

        return False

    def resolve_filter(self, query):

        def impl(dataset):

            changes = Change.select()

            # Exclude implicit changes
            if not self.include_implicit:
                changes.add_filter(Change.is_explicit_change.equal(True))

            # Filter by target / target type
            if isinstance(self.target, type):
                changes.add_filter(Change.target.one_of(self.target.select()))
            elif self.target is not None:
                changes.add_filter(Change.target.equal(self.target))

            # Filter by action
            if self.action:
                changes.add_filter(Change.action.equal(self.action))

            # Intersect with the changesets owning the matching changes
            change_index = ChangeSet.changes_index
            subset = set(
                change_index[change_id] for change_id in changes.execute()
            )
            dataset.intersection_update(subset)
            return dataset

        return ((0, 0), impl)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.events import when
from cocktail import schema
from woost.models import Item, Configuration
from woost.extensions.translationworkflow.request \
    import TranslationWorkflowRequest
from .utils import (
    get_models_included_in_translation_workflow,
    object_is_included_in_translation_workflow,
    member_is_included_in_translation_workflow
)

Item.translation_workflow_request_class = TranslationWorkflowRequest

Item.add_member(
    schema.Collection("translation_requests",
        items = "woost.extensions.translationworkflow."
                "request.TranslationWorkflowRequest",
        bidirectional = True,
        integral = True,
        editable = schema.READ_ONLY,
        backoffice_read_only_control =
            "woost.extensions.translationworkflow."
            "TranslationWorkflowRequestCollectionEditor",
        member_group = "administration"
    )
)

def get_translation_request(self, source_language, target_language):

    for request in self.translation_requests:
        if (
            request.source_language == source_language
            and request.target_language == target_language
        ):
            return request

    return None

Item.get_translation_request = get_translation_request

def require_translation_request(self, source_language, target_language):
    request = self.get_translation_request(source_language, target_language)
    if request is None:
        request = self.translation_workflow_request_class()
        request.translated_item = self
        request.source_language = source_language
        request.target_language = target_language
    return request

Item.require_translation_request = require_translation_request

REQUIRE_ALL = 1
REQUIRE_MISSING_TRANSLATIONS = 2
REQUIRE_OUTDATED_TRANSLATIONS = 3

@classmethod
def require_all_translation_requests(
    cls,
    language_paths = None,
    policy = REQUIRE_MISSING_TRANSLATIONS,
    state = None,
    verbose = False
):
    if language_paths is None:
        language_paths = Configuration.instance.translation_workflow_paths

    models = get_models_included_in_translation_workflow()

    if not models:
        return []

    if verbose:
        print "Considering the following models:"
        for model in models:
            print " - ", model.__name__

    for instance in cls.select():

        if not isinstance(instance, models):
            continue

        if verbose:
            printed_heading = False

        for source_language, target_languages in language_paths.iteritems():
            if source_language in instance.translations:
                for target_language in target_languages:
                    if policy == REQUIRE_MISSING_TRANSLATIONS:
                        if target_language in instance.translations:
                            continue
                    elif policy == REQUIRE_OUTDATED_TRANSLATIONS:
                        source_date = instance.get(
                            "last_translation_update_time",
                            source_language
                        )
                        target_date = instance.get(
                            "last_translation_update_time",
                            target_language
                        )
                        if trans_date is not None \
                        and target_date >= source_date:
                            continue

                    request = instance.require_translation_request(
                        source_language,
                        target_language
                    )

                    is_new_request = request.insert()

                    if state is not None and is_new_request:
                        request.state = state

                    if verbose and is_new_request:
                        if not printed_heading:
                            print
                            print instance
                            printed_heading = True
                        print u"  %s -> %s" % (
                            source_language,
                            target_language
                        )

Item.require_all_translation_requests = require_all_translation_requests

def invalidate_translation_request(self, source_language, target_language):
    request = self.require_translation_request(
        source_language,
        target_language
    )
    if not request.insert():
        new_state = request.state.state_after_source_change
        if new_state is not None:
            # TODO: notify the assigned translator
            # ("The source text for the translation you were working on has
            #   changed and the translation request has been pulled back")
            request.state = new_state

Item.invalidate_translation_request = invalidate_translation_request

schema.Member.included_in_translation_workflow = True
schema.SchemaObject.included_in_translation_workflow = False
Item.last_translation_update_time.included_in_translation_workflow = False

@when(Item.inserted)
def generate_translation_requests_for_new_item(e):
    if object_is_included_in_translation_workflow(e.source):
        for member in e.source.__class__.iter_members():
            if member_is_included_in_translation_workflow(member):
                language_paths = Configuration.instance.translation_workflow_paths
                for source_language in e.source.translations:
                    target_languages = language_paths.get(source_language)
                    if target_languages:
                        for target_language in target_languages:
                            request = e.source.translation_workflow_request_class()
                            request.translated_item = e.source
                            request.source_language = source_language
                            request.target_language = target_language
                            request.insert()

@when(Item.changed)
def update_translation_requests_for_modified_item(e):
    if (
        e.source.is_inserted
        and object_is_included_in_translation_workflow(e.source)
        and member_is_included_in_translation_workflow(e.member)
        and e.value
    ):
        language_paths = Configuration.instance.translation_workflow_paths
        target_languages = language_paths.get(e.language)
        if target_languages:
            for target_language in target_languages:
                e.source.invalidate_translation_request(
                    e.language,
                    target_language
                )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Configuration, LocaleMember


Configuration.add_member(
    schema.Mapping("translation_workflow_paths",
        keys = LocaleMember(
            "source_language",
            required = True
        ),
        values = schema.Collection(
            "target_languages",
            items = LocaleMember(
                required = True
            ),
            min = 1
        ),
        edit_control = "cocktail.html.TextArea",
        request_items_separator = "\n",
        member_group = "language"
    )
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Pepe Osca <pepe.osca@whads.com>
"""
from ZODB.POSException import ConflictError
from cocktail.events import Event
from cocktail import schema
from cocktail.persistence import transaction
from woost import app
from woost.models import Item, User

MAX_COMMIT_ATTEMPTS = 5


class IdentityProvider(Item):

    instantiable = False
    visible_from_root = False

    user_authenticated = Event(
        """An event triggered when the provider resolves a user.

        .. attribute:: user

            The `user <woost.models.User>` resolved by the provider.

        .. attribute:: data

            The user's profile data supplied by the provider (a dictionary with
            provider specific keys).

        .. attribute:: first_login

            Indicates if this is the first time that this user has logged in
            using this provider.
        """
    )

    members_order = [
        "title",
        "hidden",
        "debug_mode"
    ]

    title = schema.String(
        descriptive = True
    )

    hidden = schema.Boolean(
        required = True,
        default = False
    )

    debug_mode = schema.Boolean(
        required = True,
        default = False
    )

    provider_name = None
    user_data_id_field = "id"
    user_data_email_field = "email"
    user_identifier = None

    def get_auth_url(self, target_url = None):
        raise ValueError(
            "%s doesn't implement the get_auth_url() method"
            % self
        )

    def login(self, data):
        user = transaction(self.process_user_data, (data,))
        app.authentication.set_user_session(user)
        return user

    def process_user_data(self, data):

        id = data[self.user_data_id_field]
        email = data.get(self.user_data_email_field)
        user = (
            User.get_instance(**{self.user_identifier: id})
            or (
                email
                and User.get_instance(email = email)
            )
        )

        if user is None:
            user = User()

        if not user.get(self.user_identifier):
            user.set(self.user_identifier, id)
            if not user.email:
                user.email = email
            first_login = True
        else:
            first_login = False

        self.user_authenticated(
            user = user,
            data = data,
            first_login = first_login
        )
        user.insert()
        return user




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from .block import Block


class CustomBlock(Block):

    instantiable = True
    type_group = "blocks.custom"

    view_class = schema.String(
        required = True,
        shadows_attribute = True,
        before_member = "controller",
        member_group = "behavior"
    )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models.messagestyles import permission_doesnt_match_style
from woost.models.permission import ContentPermission
from .request import TranslationWorkflowRequest
from .transition import TranslationWorkflowTransition


class TranslationWorkflowTransitionPermission(ContentPermission):

    transitions = schema.Collection(
        items = schema.Reference(
            type = TranslationWorkflowTransition
        ),
        related_end = schema.Collection(),
        after_member = "content_expression"
    )

    def match(
        self,
        user,
        transition,
        translation_request = None,
        verbose = False
    ):
        if not ContentPermission.match(
            self,
            user,
            TranslationWorkflowRequest
                if translation_request is None
                else translation_request.translated_item,
            verbose = verbose
        ):
            return False

        if self.transitions and transition not in self.transitions:
            if verbose:
                print permission_doesnt_match_style(
                    "transition doesn't match"
                )
            return False

        return True




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
from cocktail import schema
from cocktail.controllers import (
    request_property,
    get_parameter,
    Pagination
)
from .block import Block
from .elementtype import ElementType
from .publishable import Publishable


class PublishableListing(Block):

    max_page_size = 50

    instantiable = True
    type_group = "blocks.listings"
    block_display = "woost.views.PublishableListingDisplay"

    groups_order = list(Block.groups_order)
    groups_order.insert(groups_order.index("content") + 1, "listing")

    members_order = [
        "element_type",
        "publishables",
        "item_accessibility",
        "listing_order",
        "links_open_in_new_window",
        "links_force_download",
        "paginated",
        "page_size",
        "view_class"
    ]

    element_type = ElementType(
        member_group = "content"
    )

    publishables = schema.Collection(
        items = schema.Reference(type = Publishable),
        related_end = schema.Collection(),
        invalidates_cache = True,
        member_group = "listing"
    )

    item_accessibility = schema.String(
        required = True,
        enumeration = [
            "accessible",
            "published",
            "any"
        ],
        default = "accessible",
        member_group = "listing"
    )

    listing_order = schema.String(
        default = "arbitrary",
        enumeration = ["arbitrary", "-last_update_time"],
        member_group = "listing"
    )

    links_open_in_new_window = schema.Boolean(
        required = True,
        default = False,
        member_group = "listing"
    )

    links_force_download = schema.Boolean(
        required = True,
        default = False,
        member_group = "listing"
    )

    paginated = schema.Boolean(
        required = True,
        default = False,
        member_group = "listing"
    )

    page_size = schema.Integer(
        min = 1,
        required = paginated,
        member_group = "listing"
    )

    view_class = schema.String(
        required = True,
        shadows_attribute = True,
        enumeration = [
            "woost.views.PublishableTextualListing",
            "woost.views.PublishableIconListing",
            "woost.views.PublishableGrid",
            "woost.views.PublishableCollage"
        ],
        default = "woost.views.PublishableTextualListing",
        member_group = "listing"
    )

    def init_view(self, view):
        Block.init_view(self, view)
        view.tag = self.element_type
        view.name_prefix = self.name_prefix
        view.name_suffix = self.name_suffix
        view.links_open_in_new_window = self.links_open_in_new_window
        view.links_force_download = self.links_force_download

        if not self.publishables:
            view.depends_on(Publishable)

        if self.paginated:
            view.pagination = self.pagination
        else:
            view.publishables = self.select_publishables()

    def select_publishables(self):

        if self.item_accessibility == "accessible":
            publishables = Publishable.select_accessible()
        elif self.item_accessibility == "published":
            publishables = Publishable.select_published()
        elif self.item_accessibility == "any":
            publishables = Publishable.select()

        publishables.base_collection = self.publishables
        self._apply_order(publishables)
        return publishables

    def _apply_order(self, publishables):
        if self.listing_order != "arbitrary":
            publishables.add_order(self.listing_order)

    @request_property
    def pagination(self):
        return get_parameter(
            self.pagination_member,
            errors = "set_default",
            prefix = self.name_prefix,
            suffix = self.name_suffix
        )

    @request_property
    def pagination_member(self):
        return Pagination.copy(**{
            "page_size.default": self.page_size,
            "page_size.max": self.max_page_size,
            "items": self.select_publishables()
        })




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.controllers import context
from woost.controllers.backoffice.editstack import StackNode
from .transition import TranslationWorkflowTransition
from .request import TranslationWorkflowRequest


class TranslationWorkflowTransitionNode(StackNode):

    requests = None
    transition = None

    def __getstate__(self):

        state = {}

        for key, value in self.__dict__.iteritems():
            if key == "requests":
                value = [request.id for request in value]
            elif key == "transition":
                value = value.id
            state[key] = value

        return state

    def __setstate__(self, state):

        for key, value in state.iteritems():
            if key == "requests":
                value = filter(None, [
                    TranslationWorkflowRequest.get_instance(id)
                    for id in value
                ])
            elif key == "transition":
                value = TranslationWorkflowTransition.get_instance(value)

            setattr(self, key, value)

    def uri(self, **params):

        if "edit_stack" not in params:
            params["edit_stack"] = self.stack.to_param(self.index)

        return context["cms"].contextual_uri(
            "translation_workflow_transition", **params
        )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import cherrypy
from simplejson import dumps
from cocktail.translations import language_context, translations
from cocktail.persistence import PersistentObject
from cocktail.controllers import Controller


class CMSMetadataController(Controller):

    def __call__(self, language, format = "json"):

        with language_context(language):

            data = {"types": {}}

            for cls in PersistentObject.schema_tree():

                if cls.translation_source is not None:
                    continue

                type_info = {
                    "name": cls.name,
                    "namespace": cls.__module__,
                    "label": translations(cls.name),
                    "plural": translations(cls.name + "-plural")
                }

                for base in cls.__bases__:
                    if issubclass(base, PersistentObject):
                        type_info["base"] = \
                            base.get_qualified_name(include_ns = True)
                        break

                full_name = cls.get_qualified_name(include_ns = True)
                data["types"][full_name] = type_info

            if format == "json":
                cherrypy.response.headers["Content-Type"] = "application/json"
                return dumps(data)
            elif format == "javascript":
                cherrypy.response.headers["Content-Type"] = "text/javascript"
                javascript = [u"cocktail.declare('woost.metadata');"]
                for key, value in data.iteritems():
                    javascript.append(
                        u"woost.metadata.%s = %s;"
                        % (key, dumps(value))
                    )
                return u"\n".join(javascript)
            else:
                raise ValueError(
                    "The 'format' parameter should be one of 'json' or "
                    "'javascript'; got %r instead" % format
                )




#-*- coding: utf-8 -*-
"""

@author:		Javier Marrero
@contact:		javier.marrero@whads.com
@organization:	Whads/Accent SL
@since:			January 2010
"""
import hashlib
import cherrypy
from cocktail import schema
from cocktail.persistence import transaction
from cocktail.controllers import request_property
from woost import app
from woost.models import User
from woost.controllers.documentcontroller import DocumentController
from woost.extensions.signup import SignUpExtension


def generate_confirmation_hash(email):
    hash = hashlib.sha1()
    hash.update(email)
    hash.update(SignUpExtension.instance.secret_key)
    return hash.hexdigest()


class SignUpConfirmationController(DocumentController):

    autologin = True

    def __init__(self, *args, **kwargs):
        self.email = self.params.read(schema.String("email"))
        self.hash = self.params.read(schema.String("hash"))

    def confirm_user(self, user):
        user.confirmed_email = True
        user.enabled = True

    def submit(self):

        if self.email or self.hash:
            # Checking hash code
            if generate_confirmation_hash(self.email) == self.hash:
                instance = User.get_instance(email=self.email)
                if instance:
                    # Confirming and enabling user instance
                    transaction(self.confirm_user, action_args = [instance])

                    # Autologin after confirmation
                    if self.autologin:
                        app.authentication.set_user_session(instance)

                    raise cherrypy.HTTPRedirect(app.publishable.get_uri())

            raise cherrypy.HTTPError(400, "Invalid hash")




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
from cocktail.translations import translations

# IssuuDocument
#------------------------------------------------------------------------------
translations.define("IssuuDocument",
    ca = u"Document Issuu",
    es = u"Documento Issuu",
    en = u"Issuu document"
)

translations.define("IssuuDocument-plural",
    ca = u"Documents Issuu",
    es = u"Documentos Issuu",
    en = u"Issuu documents"
)

translations.define("IssuuDocument.content",
    ca = u"Contingut",
    es = u"Contingut",
    en = u"Contingut"
)

translations.define("IssuuDocument.title",
    ca = u"Títol",
    es = u"Título",
    en = u"Title"
)

translations.define("IssuuDocument.issuu_document_url",
    ca = u"URL document Issuu",
    es = u"URL documento Issuu",
    en = u"Issuu document URL"
)

translations.define("IssuuDocument.thumbnail_page",
    ca = u"Pàgina de la miniatura",
    es = u"Página de la miniatura",
    en = u"Thumbnail page"
)

translations.define("IssuuDocument.issuu_document_id",
    ca = u"ID document Issuu",
    es = u"ID documento Issuu",
    en = u"Issuu document ID"
)

translations.define("IssuuDocument.issuu_config_id",
    ca = u"ID configuració Issuu",
    es = u"ID configuración Issuu",
    en = u"Issuu config ID"
)

# IssuuBlock
#------------------------------------------------------------------------------
translations.define("IssuuBlock",
    ca = u"Document Issuu",
    es = u"Documento Issuu",
    en = u"Issuu document"
)

translations.define("IssuuBlock-plural",
    ca = u"Documents Issuu",
    es = u"Documentos Issuu",
    en = u"Issuu documents"
)

translations.define("IssuuBlock.element_type",
    ca = u"Tipus d'element",
    es = u"Tipo de elemento",
    en = u"Element type"
)

translations.define("IssuuBlock.issuu_document",
    ca = u"Document Issuu",
    es = u"Documento Issuu",
    en = u"Issuu document"
)

translations.define("woost.extensions.issuu.issuu_document_controller.title",
    ca = u"Controlador de document Issuu",
    es = u"Controlador de documento Issuu",
    en = u"Issuu document controller"
)

translations.define(
    "woost.extensions.issuu.issuudocument."
    "IssuuSearchAPIError-instance",
    ca = lambda instance: u"Error consulta a la Search API: <pre>%s</pre>"
                          % instance.response,
    es = lambda instance: u"Error consulta a la Search API: <pre>%s</pre>"
                          % instance.response,
    en = lambda instance: u"Issuu search API error: <pre>%s</pre>"
                          % instance.response
)

translations.define("IssuuBlock.width",
    ca = u"Amplada",
    es = u"Ancho",
    en = u"Width"
)

translations.define("IssuuBlock.height",
    ca = u"Alçada",
    es = u"Alto",
    en = u"Height"
)

# IssuuDocumentRenderer
#------------------------------------------------------------------------------
translations.define("IssuuDocumentRenderer",
    ca = u"Pintador de document Issuu",
    es = u"Pintador de documento Issuu",
    en = u"Issuu document renderer"
)

translations.define("IssuuDocumentRenderer-plural",
    ca = u"Pintador de documents Issuu",
    es = u"Pintador de documentos Issuu",
    en = u"Issuu documents renderer"
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.stringutils import normalize
from cocktail.translations import translations
from cocktail.html import templates

List = templates.get_class("cocktail.html.List")


class TranslationsList(List):

    def _fill_entries(self):
        self.items = list(self.items)
        self.items.sort(
            key = lambda locale:
                normalize(translations("locale", locale = locale))
        )
        List._fill_entries(self)

    def create_entry_content(self, item):
        return translations("locale", locale = item)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.events import event_handler
from cocktail import schema


class Slot(schema.Collection):

    def __init__(self, *args, **kwargs):

        if "items" not in kwargs:
            from woost.models.block import Block
            kwargs["items"] = schema.Reference(type = Block)

        kwargs["related_end"] = schema.Collection()
        kwargs["cascade_delete"] = True
        kwargs.setdefault("text_search", True)
        kwargs.setdefault("cascade_cache_invalidation", "always")
        kwargs.setdefault("listable", False)
        kwargs.setdefault("searchable", False)
        kwargs.setdefault("member_group", "content")
        schema.Collection.__init__(self, *args, **kwargs)

    @event_handler
    def handle_attached(cls, e):
        if (
            getattr(e.source.schema, "cacheable", False)
            and not e.source.cache_part
        ):
            e.source.cache_part = e.source.name




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Pepe Osca <pepe.osca@whads.com>
"""
from cocktail.translations import translations

# Configuration
#------------------------------------------------------------------------------
translations.define("Configuration.identity_providers",
    ca = u"Serveis d'autenticació d'usuaris",
    es = u"Servicios de autenticación de usuarios",
    en = u"User authentication services"
)

# User
#------------------------------------------------------------------------------
translations.define("User.google_user_id",
    ca = u"Id de Google",
    es = u"Id de Google",
    en = u"Google ID"
)

translations.define("User.facebook_user_id",
    ca = u"Id de Facebook",
    es = u"Id de Facebook",
    en = u"Facebook ID"
)

# IdentityProvider
#------------------------------------------------------------------------------
translations.define("IdentityProvider",
    ca = u"Servei d'autenticació d'usuaris",
    es = u"Servicio de autenticación de usuarios",
    en = u"User authentication service"
)

translations.define("IdentityProvider-plural",
    ca = u"Serveis d'autenticació d'usuaris",
    es = u"Servicios de autenticación de usuarios",
    en = u"User authentication services"
)

translations.define("IdentityProvider.title",
    ca = u"Nom",
    es = u"Nombre",
    en = u"Name"
)

translations.define("IdentityProvider.hidden",
    ca = u"Ocult",
    es = u"Oculto",
    en = u"Hidden"
)

translations.define("IdentityProvider.debug_mode",
    ca = u"Mostrar informació per desenvolupadors",
    es = u"Mostrar información para desarrolladores",
    en = "Enable debugging"
)

# FacebookIdentityProvider
#------------------------------------------------------------------------------
translations.define("FacebookIdentityProvider",
    ca = u"Servei d'autenticació de Facebook",
    es = u"Servicio de autenticación de Facebook",
    en = u"Facebook authentication service"
)

translations.define("FacebookIdentityProvider-plural",
    ca = u"Serveis d'autenticació de Facebook",
    es = u"Servicios de autenticación de Facebook",
    en = u"Facebook authentication services"
)

translations.define("FacebookIdentityProvider.client_id",
    ca = u"Identificador de client",
    es = u"Identificador de cliente",
    en = u"Client id"
)

translations.define("FacebookIdentityProvider.client_id-explanation",
    ca = u"Un identificador únic de l'aplicació de Facebook que s'utilitzarà "
         u"per gestionar l'autenticació. S'obté des del panell de "
         u"configuració de l'aplicació de Facebook.",
    es = u"Un identificador único de la aplicación de Facebook que se "
         u"utilizará para gestionar la autenticación. Se obtiene desde el "
         u"panel de configuración de la aplicación de Facebook.",
    en = u"The unique identifier of the application that is to be used to "
         u"mediate the authentication process. Obtained from the Facebook "
         u"application configuration panel."
)

translations.define("FacebookIdentityProvider.client_secret",
    ca = u"Secret de client",
    es = u"Secreto de cliente",
    en = u"Client secret"
)

translations.define("FacebookIdentityProvider.client_secret-explanation",
    ca = u"La contrasenya de l'aplicació de Facebook que s'utilitzarà "
         u"per gestionar l'autenticació. S'obté des del panell de "
         u"configuració de l'aplicació de Facebook.",
    es = u"La contraseña de la aplicación de Facebook que se "
         u"utilizará para gestionar la autenticación. Se obtiene desde el "
         u"panel de configuración de la aplicación de Facebook.",
    en = u"The password for the application that is to be used to "
         u"mediate the authentication process. Obtained from the Facebook "
         u"application configuration panel."
)

translations.define("FacebookIdentityProvider.scope",
    ca = u"Privilegis a sol·licitar",
    es = u"Privilegios a solicitar",
    en = u"Requested scope"
)

# GoogleIdentityProvider
#------------------------------------------------------------------------------
translations.define("GoogleIdentityProvider",
    ca = u"Servei d'autenticació de Google",
    es = u"Servicio de autenticación de Google",
    en = u"Google authentication service"
)

translations.define("GoogleIdentityProvider-plural",
    ca = u"Serveis d'autenticació de Google",
    es = u"Servicios de autenticación de Google",
    en = u"Google authentication services"
)

translations.define("GoogleIdentityProvider.client_id",
    ca = u"Identificador de client",
    es = u"Identificador de cliente",
    en = u"Client id"
)

translations.define("GoogleIdentityProvider.client_id-explanation",
    ca = u"Un identificador únic de l'aplicació de Google que s'utilitzarà "
         u"per gestionar l'autenticació. S'obté a la consola de "
         u"desenvolupament de Google.",
    es = u"Un identificador único de la aplicación de Google que se "
         u"utilizará para gestionar la autenticación. Se obtiene desde la "
         u"consola de desarrollo de Google.",
    en = u"The unique identifier of the application that is to be used to "
         u"mediate the authentication process. Obtained from the Google "
         u"development console."
)

translations.define("GoogleIdentityProvider.client_secret",
    ca = u"Secret de client",
    es = u"Secreto de cliente",
    en = u"Client secret"
)

translations.define("GoogleIdentityProvider.client_secret-explanation",
    ca = u"La contrasenya de l'aplicació de Google que s'utilitzarà "
         u"per gestionar l'autenticació. S'obté a la consola de "
         u"desenvolupament de Google.",
    es = u"La contraseña de la aplicación de Google que se "
         u"utilizará para gestionar la autenticación. Se obtiene desde la "
         u"consola de desarrollo de Google.",
    en = u"The password for the application that is to be used to "
         u"mediate the authentication process. Obtained from the Google "
         u"development console."
)

translations.define("GoogleIdentityProvider.scope",
    ca = u"Privilegis a sol·licitar",
    es = u"Privilegios a solicitar",
    en = u"Requested scope"
)

translations.define("GoogleIdentityProvider.access_type",
    ca = u"Tipus d'accés",
    es = u"Tipo de acceso",
    en = u"Access type"
)

translations.define("GoogleIdentityProvider.access_type=online",
    ca = u"En línia",
    es = u"En línea",
    en = u"Online"
)

translations.define("GoogleIdentityProvider.access_type=offline",
    ca = u"Fora de línia",
    es = u"Fuera de línea",
    en = u"Offline"
)




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			February 2009
"""
from cocktail.translations import translations, get_language
from cocktail.html.element import Element
from cocktail.html import templates
from cocktail.html.utils import rendering_xml
from cocktail.controllers import context
from woost import app
from woost.models import (
    Configuration,
    ReadTranslationPermission
)

LinkSelector = templates.get_class("cocktail.html.LinkSelector")


class LanguageSelector(LinkSelector):

    tag = "ul"
    translated_labels = True
    missing_translations = "redirect" # "redirect", "hide", "disable"
    autohide = True

    def create_entry(self, item):

        entry = Element("li")
        link = self.create_entry_link(item)

        if self.is_selected(item):
            strong = Element("strong")
            strong.append(link)
            entry.append(strong)
        else:
            entry.append(link)

        if self.missing_translations != "redirect":
            publishable = app.publishable
            value = self.get_item_value(item)
            if not publishable.is_accessible(language = value):
                if self.missing_translations == "hide":
                    entry.visible = False
                elif self.missing_translations == "disable":
                    link.tag = "span"
                    link["href"] = None

        return entry

    def get_languages(self):
        config = Configuration.instance
        user = app.user
        return [
            language
            for language in (
                config.get_setting("published_languages")
                or config.languages
            )
            if user.has_permission(
                ReadTranslationPermission,
                language = language
            )
        ]

    def _ready(self):

        if self.items is None:
            self.items = self.get_languages()

        if self.value is None:
            self.value = get_language()

        if self.autohide and len(self.items) < 2:
            self.visible = False
        else:
            LinkSelector._ready(self)

            # Hack for IE <= 6
            if self.children:
                self.children[-1].add_class("last")

    def get_item_label(self, language):
        if self.translated_labels:
            return translations("locale", locale = language, language = language)
        else:
            return translations("locale", locale = language)

    def get_entry_url(self, language):
        cms = context["cms"]
        publishable = app.publishable

        if self.missing_translations == "redirect" \
        and not publishable.is_accessible(language = language):
            path = "/"
        else:
            path = None

        return cms.translate_uri(path = path, language = language)

    def create_entry_link(self, item):
        link = LinkSelector.create_entry_link(self, item)
        value = self.get_item_value(item)
        link["lang"] = value
        link["hreflang"] = value

        @link.when_ready
        def set_xml_lang():
            if rendering_xml():
                link["xml:lang"] = value

        return link




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
from cocktail import schema
from cocktail.translations import translations
from woost.models import Website, Publishable
from woost.extensions.ecommerce.ecommercebillingconcept \
    import ECommerceBillingConcept
from woost.extensions.payments import PaymentsExtension
from woost.extensions.payments.paymentgateway import PaymentGateway


Website.add_member(
    schema.Collection(
        "ecommerce_payment_types",
        items = schema.String(
            enumeration = ("payment_gateway", "transfer", "cash_on_delivery"),
            translate_value = lambda value, language = None, **kwargs:
                translations(
                    "Website.ecommerce_payment_types=%s" % value,
                    language = language
                ),
        ),
        min = 1,
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Collection(
        "ecommerce_pricing",
        items = schema.Reference(type = ECommerceBillingConcept),
        related_end = schema.Collection(),
        integral = True,
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Collection(
        "ecommerce_shipping_costs",
        items = schema.Reference(type = ECommerceBillingConcept),
        related_end = schema.Collection(),
        integral = True,
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Collection(
        "ecommerce_taxes",
        items = schema.Reference(type = ECommerceBillingConcept),
        related_end = schema.Collection(),
        integral = True,
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Reference(
        "ecommerce_default_catalog",
        type = Publishable,
        related_end = schema.Collection(),
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Collection(
        "ecommerce_order_steps",
        items = schema.Reference(type = Publishable),
        related_end = schema.Collection(),
        listed_by_default = False,
        member_group = "ecommerce"
    )
)

Website.add_member(
    schema.Collection(
        "ecommerce_orders",
        items = "woost.extensions.ecommerce.ecommerceorder."
                "ECommerceOrder",
        bidirectional = True,
        visible = False,
        member_group = "ecommerce"
    )
)

Website.members_order += [
    "ecommerce_payment_types",
    "ecommerce_pricing",
    "ecommerce_shipping_costs",
    "ecommerce_taxes",
    "ecommerce_default_catalog",
    "ecommerce_order_steps"
]

payments_ext = PaymentsExtension.instance

if payments_ext.enabled:
    Website.add_member(
        schema.Reference(
            "ecommerce_payment_gateway",
            type = PaymentGateway,
            required = True,
            related_end = schema.Collection(),
            listed_by_default = False,
            member_group = "ecommerce"
        )
    )

    Website.members_order += [
        "ecommerce_payment_gateway"
    ]




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Pepe Osca <pepe.osca@whads.com>
"""
import cherrypy
from oauth2client.client import OAuth2WebServerFlow
import httplib2
from googleapiclient import discovery
from cocktail.styled import styled
from cocktail.controllers import Controller, Location, session
from woost import app
from woost.controllers import CMSController
from .identityprovider import GoogleIdentityProvider

SESSION_PREFIX = "woost.extensions.identity.google."


class GoogleOAuthController(Controller):

    def resolve(self, path):

        if not path:
            raise cherrypy.HTTPError(400)

        provider_id = path[0]

        try:
            provider_id = int(provider_id)
        except ValueError:
            raise cherrypy.HTTPError(400)

        provider = GoogleIdentityProvider.get_instance(provider_id)
        if provider is None:
            raise cherrypy.NotFound()

        path.pop(0)
        return GoogleOAuthProviderController(provider)


class GoogleOAuthProviderController(Controller):

    def __init__(self, provider):
        Controller.__init__(self)
        self.provider = provider
        self.target_url = (
            session.get(SESSION_PREFIX + "target_url")
            or app.website.home.get_uri()
        )

    def step_url(self, step_number):
        location = Location.get_current_host()
        location.path_info = "/google_oauth/%d/step%d" % (
            self.provider.id,
            step_number
        )
        return unicode(location).decode("utf-8")

    @cherrypy.expose
    def step1(self, code = None, target_url = None, **kwargs):

        if target_url:
            self.target_url = target_url
            session[SESSION_PREFIX + "target_url"] = target_url

        flow = OAuth2WebServerFlow(
            self.provider.client_id,
            self.provider.client_secret,
            self.provider.scope,
            redirect_uri = self.step_url(1)
        )
        flow.params["access_type"] = self.provider.access_type

        if not code:
            raise cherrypy.HTTPRedirect(flow.step1_get_authorize_url())

        if self.provider.debug_mode:
            print styled("Google authorization code:", "magenta"), code

        credentials = flow.step2_exchange(code)
        session[SESSION_PREFIX + "credentials"] = credentials

        if self.provider.debug_mode:
            print styled("Google refresh token:", "magenta"),
            print credentials.refresh_token
            print styled("Google access token:", "magenta"),
            print credentials.access_token

        raise cherrypy.HTTPRedirect(self.step_url(2))

    @cherrypy.expose
    def step2(self):

        credentials = session.get(SESSION_PREFIX + "credentials")

        if not credentials or credentials.access_token_expired:
            raise cherrypy.HTTPRedirect(self.step_url(1))

        http_auth = credentials.authorize(httplib2.Http())
        oauth2_service = discovery.build('oauth2', 'v2', http_auth)
        user_data = oauth2_service.userinfo().get().execute()

        if self.provider.debug_mode:
            print styled("Google user profile:", "magenta"), user_data

        self.provider.login(user_data)
        del session[SESSION_PREFIX + "credentials"]

        raise cherrypy.HTTPRedirect(self.target_url)

    @cherrypy.expose
    def refresh_token(self, code):
        response = self.provider.get_refresh_token(code)
        return (
            """
            <input id="code" type="text" autofocus value="%s">
            <script type="text/javascript">
                window.onload = function () {
                    document.getElementById("code").select();
                }
            </script>
            """
            % response["refresh_token"]
        )


CMSController.google_oauth = GoogleOAuthController




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations

translations.define(
    "woost.extensions.googleanalytics."
    "default_custom_definitions.locale.title",
    ca = u"Woost: idioma",
    es = u"Woost: idioma",
    en = u"Woost: locale"
)

translations.define(
    "woost.extensions.googleanalytics."
    "default_custom_definitions.roles.title",
    ca = u"Woost: rols d'usuari",
    es = u"Woost: roles de usuario",
    en = u"Woost: user roles"
)

translations.define(
    "woost.extensions.googleanalytics."
    "default_custom_definitions.path.title",
    ca = u"Woost: camí",
    es = u"Woost: camino",
    en = u"Woost: path"
)

translations.define(
    "woost.extensions.googleanalytics."
    "default_custom_definitions.publishable.title",
    ca = u"Woost: publicable",
    es = u"Woost: publicable",
    en = u"Woost: publishable"
)

translations.define(
    "woost.extensions.googleanalytics."
    "default_custom_definitions.type.title",
    ca = u"Woost: tipus",
    es = u"Woost: tipo",
    en = u"Woost: type"
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from cocktail import schema
from woost.models import User, LocaleMember
from .path import TranslationWorkflowPath
from .translationagency import TranslationAgency

User.members_order += [
    "managed_translation_agency",
    "translation_agency",
    "translation_proficiencies"
]

User.add_member(
    schema.Reference("managed_translation_agency",
        type = TranslationAgency,
        bidirectional = True,
        related_key = "managers",
        editable = schema.READ_ONLY,
        listed_by_default = False
    )
)

User.add_member(
    schema.Reference("translation_agency",
        type = TranslationAgency,
        bidirectional = True,
        related_key = "translators",
        editable = schema.READ_ONLY,
        indexed = True,
        listed_by_default = False
    )
)

User.add_member(
    schema.Collection("translation_proficiencies",
        items = TranslationWorkflowPath
    )
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
import logging
import smtplib
from email.mime.text import MIMEText
from email.Header import Header
from email.Utils import formatdate
from cocktail import schema
from cocktail.asynctask import TaskManager
from cocktail.controllers import context
from cocktail.translations import translations, set_language
from cocktail.persistence import datastore
from cocktail.schema.exceptions import ValidationError
from woost import app
from woost.models import (
    Configuration,
    Item,
    User
)
from woost.extensions.mailer.mailinglist import MailingList
from woost.extensions.mailer.sendemailpermission import SendEmailPermission
from woost.extensions.mailer.htmladapter import HTMLAdapter


logger = logging.getLogger("woost.extensions.mailer")
tasks = TaskManager()

MAILING_STARTED = 1
MAILING_FINISHED = 2

class RunningMailingError(Exception):
    """An exception raised when trying to delete a running mailing"""


class DocumentTemplateRequiredError(ValidationError):

    def __str__(self):
        return "%s (Document must have a template)" \
                % ValidationError.__str__(self)

class LanguageValueError(ValidationError):

    def __str__(self):
        return "%s (Language isn't one of the document languages)" \
                % ValidationError.__str__(self)

class PerUserCustomizableValueError(ValidationError):

    def __str__(self):
        return "%s (Document template isn't per user customizable)" \
            % ValidationError.__str__(self)


def available_lists(ctx):
    user = app.user
    return [mailingList
            for mailingList in MailingList.select()
            if user.has_permission(
                SendEmailPermission, mailingList = mailingList
            )
            and len(get_receivers_by_lists([mailingList]))]


class Mailing(Item):

    emails_per_transaction = 10
    __cached_body = None

    # Members
    #------------------------------------------------------------------------------
    members_order = [
        "sender", "subject", "document", "language", "per_user_customizable",
        "lists"
    ]
    groups_order = "content", "administration"

    document = schema.Reference(
        required = True,
        type = "woost.models.document.Document",
        related_end = schema.Collection(),
        member_group = "content"
    )

    sender = schema.String(
        required = True,
        format = r"([\w\-\.]+@(\w[\w\-]+\.)+[\w\-]+)",
        member_group = "content"
    )

    subject = schema.String(
        required = True,
        spellcheck = True,
        member_group = "content"
    )

    language = schema.String(
        required = True,
        enumeration = lambda ctx: Configuration.instance.languages,
        translate_value = lambda value, language = None, **kwargs:
            "" if not value else translations(value, **kwargs),
        edit_control = "cocktail.html.RadioSelector",
        member_group = "content"
    )

    lists = schema.Collection(
        items = schema.Reference(
            type = "woost.extensions.mailer.mailinglist.MailingList",
            default_order = "title",
            enumeration = available_lists,
            translate_value = lambda value, language = None, **kwargs: \
                "%s (%d %s)" % (
                    translations(value),
                    len(get_receivers_by_lists([value])),
                    translations("woost.extensions.mailer users")
                )
        ),
        edit_control = "cocktail.html.CheckList",
        bidirectional = True,
        min = 1,
        member_group = "content"
    )

    per_user_customizable = schema.Boolean(
        "per_user_customizable",
        default = False,
        member_group = "content"
    )

    pending = schema.Mapping(visible = False)

    sent = schema.Mapping(visible = False)

    errors = schema.Mapping(visible = False)

    total = schema.Integer(visible = False)

    status = schema.Integer(
        editable = schema.READ_ONLY,
        enumeration = (MAILING_STARTED, MAILING_FINISHED),
        translate_value = lambda value, language = None, **kwargs: \
            "" if value is None else translations(
                "woost.extensions.mailer.mailing.Mailing.status %d" % value
            ),
        member_group = "administration"
    )

    def __init__(self, *args, **kwargs):
        Item.__init__(self, *args, **kwargs)
        self._v_template_values = {}

    def delete(self, deleted_objects = None):
        if self.id in tasks and not tasks[self.id].completed:
            raise RunningMailingError("Can't delete a running mailing")

        Item.delete(self, deleted_objects)

    def _get_message(self, receiver):
        message = MIMEText(
            self.render_body(receiver),
            _subtype = self.document.mime_type.split("/")[1],
            _charset = self.document.encoding
        )
        message["To"] = receiver.email
        message["From"] = self.sender
        message["Subject"] = Header(self.subject, self.document.encoding)
        message["Date"] = formatdate()

        return message

    def render_body(self, receiver):

        if not self.per_user_customizable and self.__cached_body:
            return self.__cached_body

        values = self._v_template_values

        if self.per_user_customizable:
            values = values.copy()
            user = receiver
        else:
            user = User.require_instance(qname = "woost.anonymous_user")

        # Update context
        context["show_user_controls"] = False
        context["email_version"] = True
        values.setdefault("document_media", "email")

        current_user = app.user
        try:
            app.user = user
            body = self.document.render(**values)
            body = self._adapt_html(body)
        finally:
            app.user = current_user

        if not self.per_user_customizable:
            self.__cached_body = body

        return body

    def _adapt_html(self, html):
        adapter = HTMLAdapter(html)
        return adapter.adapt()

    def send_message(self, smtp_server, receiver):
        message = self._get_message(receiver)
        config = Configuration.instance
        try:
            return smtp_server.sendmail(self.sender, [receiver.email], message.as_string())
        except smtplib.SMTPServerDisconnected, e:
            logger.info("%d - Server disconnected, reconnecting - %s" % (
                self.id, e
            ))
            # smtplib bug
            # SMTP.quit() doesn't clear the HELO/EHLO
            # attributes, so on the next connection these
            # commands weren't sent.
            # http://bugs.python.org/issue4142
            smtp_server.helo_resp = None
            smtp_server.ehlo_resp = None
            smtp_server.connect(config.smtp_host, smtplib.SMTP_PORT)
            return smtp_server.sendmail(self.sender, [receiver.email], message.as_string())
        except smtplib.SMTPSenderRefused, e:
            logger.info("%d - Maximum number of messages per connection reached, reconnecting - %s" % (
                self.id, e
            ))
            smtp_server.quit()
            # smtplib bug
            # SMTP.quit() doesn't clear the HELO/EHLO
            # attributes, so on the next connection these
            # commands weren't sent.
            # http://bugs.python.org/issue4142
            smtp_server.helo_resp = None
            smtp_server.ehlo_resp = None
            smtp_server.connect(config.smtp_host, smtplib.SMTP_PORT)
            return smtp_server.sendmail(self.sender, [receiver.email], message.as_string())

    def send(self, smtp_server, template_values, current_context):

        if not self.id in tasks:
            current_user = app.user
            current_website = app.website
            if self.status is None:
                self.pending = self.get_receivers()
                self.total = len(self.pending)
                self.sent = {}
                self.errors = {}
                datastore.commit()

            def process():
                mailing = Mailing.get_instance(self.id)
                mailing._v_template_values = template_values.copy()
                logger.info("%d - Mailing started by user %s (%s)" % (
                    mailing.id, current_user, current_user.email
                ))

                set_language(mailing.language)
                app.website = current_website
                context.update(current_context)
                mailing.status = MAILING_STARTED
                processed_emails = 0
                try:
                    for email, receiver in mailing.pending.items():
                        try:
                            mailing.send_message(smtp_server, receiver)
                        except Exception, e:
                            logger.exception("%d - %s (%s) - %s" % (
                                mailing.id, receiver, email, e
                            ))
                            mailing.errors[email] = e
                        else:
                            logger.info("%d - Email sent to %s (%s)" % (
                                mailing.id, receiver, email
                            ))
                            mailing.sent[email] = receiver
                            del mailing.pending[email]
                            try:
                                del mailing.errors[email]
                                logger.info("%d - Removed %s (%s) from errors" % (
                                    mailing.id, receiver, email
                                ))
                            except KeyError:
                                pass
                        finally:
                            processed_emails += 1
                            if processed_emails >= mailing.emails_per_transaction:
                                mailing._p_changed = True
                                datastore.commit()
                                processed_emails = 0
                finally:
                    if not mailing.pending:
                        mailing.status = MAILING_FINISHED
                    mailing._p_changed = True
                    try:
                        datastore.commit()
                    finally:
                        try:
                            smtp_server.quit()
                        finally:
                            datastore.close()

            task = tasks.task(process, id = self.id)
            task.mailing_id = self.id
            task.user_id = current_user.id

    def get_receivers(self):
        return get_receivers_by_lists(self.lists)


def get_receivers_by_lists(lists):
    receivers = {}
    for mailingList in lists:
        for user in mailingList.users:
            if user.enabled and getattr(user, "confirmed_email", True):
                receivers.setdefault(user.email, user)

    return receivers


# Validations
#------------------------------------------------------------------------------

def document_validation(context):
    if context.value and not context.value.template:
        yield DocumentTemplateRequiredError(context)

def language_validation(context):
    language = context.value
    document = context.get_value("document")

    if language and document and language not in document.translations.keys():
        yield LanguageValueError(context)

Mailing.document.add_validation(document_validation)
Mailing.language.add_validation(language_validation)

